#!/usr/bin/python
# -*- coding: utf-8 -*-

import json
from generator import Je, CellType


class CarControllerAutomaton(object):
    def _accel_name(self, ax, ay):
        sax = str(ax)
        say = str(ay)
        #if ax < 0:
        #    sax = "_%d" % (ax * -1)
        #if ay < 0:
        #    say = "_%d" % (ay * -1)
        return "accel%d_%s_%s" % (self.carid, sax, say)


    def __init__(self, grid, carid):
        self.grid = grid
        self.carid = carid
        self.carx = "car%d_x" % carid
        self.cary = "car%d_y" % carid
        self.cardx = "car%d_dx" % carid
        self.cardy = "car%d_dy" % carid

        self._variables = []
        self._initial_values = {}
        self._locations = []
        self._edges = []
        self._locations = []
        self._initial_loc = None



        self._locations = [
            "send_acceleration",
            "sent_acceleration",

            "branch_mode",

            "check_current_velocity",
            "wait_result_current_velocity",

            "current_velocity_invalid",
            "await_decel_x",
            "check_decel_y",
            "await_decel_y",

            "current_velocity_valid",

            "escape_wall_x_unknown",
            "escape_wall_y_unknown",
            "escape_wall_x",
            "escape_wall_y",
            "escape_wall_x_check",
            "escape_wall_y_check",

            "determine_accel_x",
            "check_accel_x",
            "determine_accel_y",
            "check_accel_y",
        ]

        self._initial_loc = "branch_mode"


        LEAVE_WALL_MODE_AFTER = 2

        # store chosen acceleration in local ax and ay
        self._variables += [
            { "name": "wall_escape_mode", "type": {"kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": 4, "initial-value": 0 } },
            { "name": "leave_wall_mode", "type": {"kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": LEAVE_WALL_MODE_AFTER, "initial-value": 0 } },
            { "name": "ax", "type": { "kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": 1, "initial-value": 0 } },
            { "name": "ay", "type": { "kind": "bounded", "base": "int", "lower-bound": -1, "upper-bound": 1, "initial-value": 0 } },
            { "name": "dx", "type":  { "kind": "bounded", "base": "int", "lower-bound": -self.grid.dimx, "upper-bound": self.grid.dimx, "initial-value": 0 } },
            { "name": "dy", "type":  { "kind": "bounded", "base": "int", "lower-bound": -self.grid.dimy, "upper-bound": self.grid.dimy, "initial-value": 0 } },
            { "name": "dxm", "type":  { "kind": "bounded", "base": "int", "lower-bound": -self.grid.dimx, "upper-bound": self.grid.dimx, "initial-value": 0 } },
            { "name": "dym", "type":  { "kind": "bounded", "base": "int", "lower-bound": -self.grid.dimy, "upper-bound": self.grid.dimy, "initial-value": 0 } },
            { "name": "dxp", "type":  { "kind": "bounded", "base": "int", "lower-bound": -self.grid.dimx, "upper-bound": self.grid.dimx, "initial-value": 0 } },
            { "name": "dyp", "type":  { "kind": "bounded", "base": "int", "lower-bound": -self.grid.dimy, "upper-bound": self.grid.dimy, "initial-value": 0 } },
        ]
        self._initial_values["wall_escape_mode"] = 0
        self._initial_values["leave_wall_mode"] = 0
        self._initial_values["ax"] = 0
        self._initial_values["ay"] = 0
        self._initial_values["dx"] = 0
        self._initial_values["dy"] = 0
        self._initial_values["dxm"] = 0
        self._initial_values["dym"] = 0
        self._initial_values["dxp"] = 0
        self._initial_values["dyp"] = 0

            

        ##################################################
        ### Branch between different modes
        self._edges += [
        {
            "location": "branch_mode",
            "guard": { "exp": Je.And(Je.Not("car%d_crashed" % self.carid), Je.Eq("wall_escape_mode", 0)) },
            "destinations": [
                { "location": "check_current_velocity" },
            ],
        },
        {
            "location": "branch_mode",
            "guard": { "exp": Je.And(Je.Not("car%d_crashed" % self.carid), Je.Or(Je.Eq("wall_escape_mode", 1), Je.Eq("wall_escape_mode", 2))) },
            "destinations": [
                { "location": "escape_wall_x" },
            ],
        },
        {
            "location": "branch_mode",
            "guard": { "exp": Je.And(Je.Not("car%d_crashed" % self.carid), Je.Or(Je.Eq("wall_escape_mode", 3), Je.Eq("wall_escape_mode", 4))) },
            "destinations": [
                { "location": "escape_wall_y" },
            ],
        }
        ]

    

        ##################################################
        ## Acceleration computed -> synchronize via corresponding action
        self._edges += [ {
            "location": "send_acceleration",
            "action": self._accel_name(ax, ay),
            "guard": { "exp": Je.And(Je.Eq("ax", ax), Je.Eq("ay", ay), Je.Or(Je.Ne(self.cardx, 0), Je.Ne(self.cardy, 0), Je.Ne("ax", 0), Je.Ne("ay", 0))) },
            "destinations": [{
                "location": "sent_acceleration",
                "assignments": [
                    Je.Value("ax", 0),
                    Je.Value("ay", 0),
                ],
            }]
        } for ax in [-1, 0, 1] for ay in [-1, 0, 1]]

        self._edges.append({
            "location": "sent_acceleration",
            "action": "synchronize",
            "guard": { "exp": Je.And(Je.Gt("wall_escape_mode", 0), Je.Ge("leave_wall_mode", 1)) },
            "destinations": [{
                "location": "branch_mode",
            }]
        })

        self._edges.append({
            "location": "sent_acceleration",
            "action": "synchronize",
            "guard": { "exp": Je.Or(Je.Eq("wall_escape_mode", 0), Je.Eq("leave_wall_mode", 0)) },
            "destinations": [{
                "location": "branch_mode",
                "assignments": [
                    Je.Value("wall_escape_mode", 0),
                    Je.Value("leave_wall_mode", 0),
                ],
            }]
        })

        ## If ax == 0 and ay == 0, but not yet at goal, moving directly to
        ## closest goal not possible due to hitting walls
        ## -> try to escape the wall
        self._edges += [
        {
            "location": "send_acceleration",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Eq("ax", 0), Je.Eq("ay", 0), Je.Eq(self.cardx, 0), Je.Eq(self.cardy, 0), Je.Ne(self.carx, "goalx")) },
            "destinations": [{
                "location": "escape_wall_x_unknown",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", 0),
                    Je.Value("map_dy", 1),
                ],
            }]
        },
        {
            "location": "escape_wall_x_unknown",
            "action": "is_valid",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", 0),
                    Je.Value("ay", 1),
                    Je.Value("wall_escape_mode", 1),
                    Je.Value("leave_wall_mode", LEAVE_WALL_MODE_AFTER),
                ],
            }]
        },
        {
            "location": "escape_wall_x_unknown",
            "action": "passed_finish_line",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", 0),
                    Je.Value("ay", 1),
                ],
            }]
        },
        {
            "location": "escape_wall_x_unknown",
            "action": "is_invalid",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", 0),
                    Je.Value("ay", -1),
                    Je.Value("wall_escape_mode", 2),
                    Je.Value("leave_wall_mode", LEAVE_WALL_MODE_AFTER),
                ],
            }]
        },

        {
            "location": "send_acceleration",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Eq("ax", 0), Je.Eq("ay", 0), Je.Eq(self.cardx, 0), Je.Eq(self.cardy, 0), Je.Eq(self.carx, "goalx")) },
            "destinations": [{
                "location": "escape_wall_y_unknown",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", 1),
                    Je.Value("map_dy", 0),
                ],
            }]
        },
        {
            "location": "escape_wall_y_unknown",
            "action": "is_valid",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", 1),
                    Je.Value("ay", 0),
                    Je.Value("wall_escape_mode", 3),
                    Je.Value("leave_wall_mode", LEAVE_WALL_MODE_AFTER),
                ],
            }]
        },
        {
            "location": "escape_wall_y_unknown",
            "action": "passed_finish_line",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", 1),
                    Je.Value("ay", 0),
                ],
            }]
        },
        {
            "location": "escape_wall_y_unknown",
            "action": "is_invalid",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", -1),
                    Je.Value("ay", 0),
                    Je.Value("wall_escape_mode", 4),
                    Je.Value("leave_wall_mode", LEAVE_WALL_MODE_AFTER),
                ],
            }]
        },
        ]


        #### Escaping walls...

        self._edges += [
        {
            "location": "escape_wall_x",
            "guard": { "exp":  Je.Ne(self.cardx, 0) },
            "action": "check_if_valid",
            "destinations": [{
                "location": "escape_wall_x_check",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Add("ax", self.cardx)),
                    Je.Value("map_dy", self.cardy),
                ],
            }]
        },
        {
            "location": "escape_wall_x",
            "guard": { "exp": Je.And(Je.Eq(self.cardx, 0), Je.Eq("ax", 0))  },
            "action": "check_if_valid",
            "destinations": [{
                "location": "escape_wall_x_check",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Ite(Je.Le(self.carx, "goalx"), 1, -1)),
                    Je.Value("map_dy", self.cardy),
                    Je.Value("ax", Je.Ite(Je.Le(self.carx, "goalx"), 1, -1)),
                ],
            }]
        },
        {
            "location": "escape_wall_x",
            "guard": { "exp":  Je.And(Je.Eq(self.cardx, 0), Je.Ne("ax", 0)) },
            "action": "check_if_valid",
            "destinations": [{
                "location": "escape_wall_x_check",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", self.cardx),
                    Je.Value("map_dy", self.cardy),
                    Je.Value("ax", 0),
                ],
            }]
        },
        {
            "location": "escape_wall_x_check",
            "action": "passed_finish_line",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("wall_escape_mode", 0),
                    Je.Value("leave_wall_mode", 0),
                ]
            }]
        },
        {
            "location": "escape_wall_x_check",
            "action": "is_valid",
            "guard": { "exp": Je.Ne(Je.Add(self.cardx, "ax"), 0) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("leave_wall_mode", Je.Sub("leave_wall_mode", 1)),
                ],
            }]
        },
        {
            "location": "escape_wall_x_check",
            "action": "is_valid",
            "guard": { "exp": Je.Eq(Je.Add(self.cardx, "ax"), 0) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("leave_wall_mode", LEAVE_WALL_MODE_AFTER),
                ],
            }]
        },
        {
            "location": "escape_wall_x_check",
            "action": "is_invalid",
            "guard": { "exp": Je.Eq(Je.Add(self.cardx, "ax"), 0) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ay", Je.Neg(self.cardy)),
                    Je.Value("wall_escape_mode", 0),
                    Je.Value("leave_wall_mode", 0),
                ],
            }]
        },
        {
            "location": "escape_wall_x_check",
            "action": "is_invalid",
            "guard": { "exp": Je.And(Je.Ne(Je.Add(self.cardx, "ax"), 0), Je.Eq("ax", 0)) },
            "destinations": [{
                "location": "escape_wall_x",
                "assignments": [
                    Je.Value("ax", Je.Neg(self.cardx)),
                ],
            }]
        },
        {
            "location": "escape_wall_x_check",
            "action": "is_invalid",
            "guard": { "exp": Je.And(Je.Ne(Je.Add(self.cardx, "ax"), 0), Je.Ne("ax", 0)) },
            "destinations": [{
                "location": "escape_wall_x",
                "assignments": [
                ],
            }]
        },
        ]

        self._edges += [
        {
            "location": "escape_wall_y",
            "guard": { "exp":  Je.Ne(self.cardy, 0) },
            "action": "check_if_valid",
            "destinations": [{
                "location": "escape_wall_y_check",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", self.cardx),
                    Je.Value("map_dy", Je.Add("ay", self.cardy)),
                ],
            }]
        },
        {
            "location": "escape_wall_y",
            "guard": { "exp": Je.And(Je.Eq(self.cardy, 0), Je.Eq("ay", 0))  },
            "action": "check_if_valid",
            "destinations": [{
                "location": "escape_wall_y_check",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", self.cardx),
                    Je.Value("map_dy", Je.Ite(Je.Le(self.cary, "goaly"), 1, -1)),
                    Je.Value("ay", Je.Ite(Je.Le(self.cary, "goaly"), 1, -1)),
                ],
            }]
        },
        {
            "location": "escape_wall_y",
            "guard": { "exp":  Je.And(Je.Eq(self.cardy, 0), Je.Ne("ay", 0)) },
            "action": "check_if_valid",
            "destinations": [{
                "location": "escape_wall_y_check",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", self.cardx),
                    Je.Value("map_dy", self.cardy),
                    Je.Value("ay", 0),
                ],
            }]
        },
        {
            "location": "escape_wall_y_check",
            "action": "passed_finish_line",
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("wall_escape_mode", 0),
                    Je.Value("leave_wall_mode", 0),
                ]
            }]
        },
        {
            "location": "escape_wall_y_check",
            "action": "is_valid",
            "guard": { "exp": Je.Ne(Je.Add(self.cardy, "ay"), 0) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("leave_wall_mode", Je.Sub("leave_wall_mode", 1)),
                ],
            }]
        },
        {
            "location": "escape_wall_y_check",
            "action": "is_valid",
            "guard": { "exp": Je.Eq(Je.Add(self.cardy, "ay"), 0) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("leave_wall_mode", LEAVE_WALL_MODE_AFTER),
                ],
            }]
        },
        {
            "location": "escape_wall_y_check",
            "action": "is_invalid",
            "guard": { "exp": Je.Eq(Je.Add(self.cardy, "ay"), 0) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", Je.Neg(self.cardx)),
                    Je.Value("wall_escape_mode", 0),
                    Je.Value("leave_wall_mode", 0),
                ],
            }]
        },
        {
            "location": "escape_wall_y_check",
            "action": "is_invalid",
            "guard": { "exp": Je.And(Je.Ne(Je.Add(self.cardy, "ay"), 0), Je.Eq("ay", 0)) },
            "destinations": [{
                "location": "escape_wall_y",
                "assignments": [
                    Je.Value("ay", Je.Neg(self.cardy)),
                ],
            }]
        },
        {
            "location": "escape_wall_y_check",
            "action": "is_invalid",
            "guard": { "exp": Je.And(Je.Ne(Je.Add(self.cardy, "ay"), 0), Je.Ne("ay", 0)) },
            "destinations": [{
                "location": "escape_wall_y",
                "assignments": [
                ],
            }]
        },
        ]


        ##############################################
        ## Check current velocity

        self._edges += [
        {
            "location": "check_current_velocity",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Ge(self.cardx, 0), Je.Ge(self.cardy, 0)) },
            "destinations": [{
                "location": "wait_result_current_velocity",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Add(self.cardx, 1)), 2))),
                    Je.Value("map_dy", Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Add(self.cardy, 1)), 2))),
                ],
            }]
        },
        {
            "location": "check_current_velocity",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Ge(self.cardx, 0), Je.Lt(self.cardy, 0)) },
            "destinations": [{
                "location": "wait_result_current_velocity",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Min(self.grid.dimx, Je.Mult(self.cardx, self.cardx))),
                    Je.Value("map_dy", Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Sub(self.cardy, 1)), 2)))),
                ],
            }]
        },
        {
            "location": "check_current_velocity",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Lt(self.cardx, 0), Je.Ge(self.cardy, 0)) },
            "destinations": [{
                "location": "wait_result_current_velocity",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Sub(self.cardx, 1)), 2)))),
                    Je.Value("map_dy", Je.Min(self.grid.dimy, Je.Mult(self.cardy, self.cardy))),
                ],
            }]
        },
        {
            "location": "check_current_velocity",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Lt(self.cardx, 0), Je.Lt(self.cardy, 0)) },
            "destinations": [{
                "location": "wait_result_current_velocity",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Sub(self.cardx, 1)), 2)))),
                    Je.Value("map_dy", Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Sub(self.cardy, 1)), 2)))),
                ],
            }]
        }]

        ### PASSED GOAL => JUST KEEP CURRENT VELOCITY (ax and ay are already set to 0)
        self._edges.append({
            "location": "wait_result_current_velocity",
            "action": "passed_finish_line",
            "destinations": [{
                "location": "send_acceleration",
            }]
        })

        ### INVALID => BREAK
        self._edges.append({
            "location": "wait_result_current_velocity",
            "action": "is_invalid",
            "destinations": [{
                "location": "current_velocity_invalid",
            }]
        })

        ### First break in x direction
        self._edges += [
        {
            "location": "current_velocity_invalid",
            "guard": { "exp": Je.Eq(self.cardx, 0) },
            "destinations": [{
                "location": "check_decel_y",
            }]
        },
        {
            "location": "current_velocity_invalid",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Gt(self.cardx, 0), Je.Ge(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_x",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Sub(self.cardx, 1)), 2))),
                    Je.Value("map_dy", Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Add(self.cardy, 1)), 2))),
                    Je.Value("ax", -1),
                ],
            }]
        },
        {
            "location": "current_velocity_invalid",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Gt(self.cardx, 0), Je.Lt(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_x",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Sub(self.cardx, 1)), 2))),
                    Je.Value("map_dy", Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Sub(self.cardy, 1)), 2)))),
                    Je.Value("ax", -1),
                ],
            }]
        },
        {
            "location": "current_velocity_invalid",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Lt(self.cardx, 0), Je.Ge(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_x",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Add(self.cardx, 1)), 2)))),
                    Je.Value("map_dy", Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Add(self.cardy, 1)), 2))),
                    Je.Value("ax", 1),
                ],
            }]
        },
        {
            "location": "current_velocity_invalid",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Lt(self.cardx, 0), Je.Lt(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_x",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Add(self.cardx, 1)), 2)))),
                    Je.Value("map_dy", Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Sub(self.cardy, 1)), 2)))),
                    Je.Value("ax", 1),
                ],
            }]
        },
        {
            "location": "await_decel_x",
            "action": "is_valid",
            "destinations": [{
                "location": "send_acceleration",
            }]
        },
        {
            "location": "await_decel_x",
            "action": "passed_finish_line",
            "destinations": [{
                "location": "send_acceleration",
            }]
        },
        {
            "location": "await_decel_x",
            "action": "is_invalid",
            "destinations": [{
                "location": "check_decel_y",
            }]
        },
        ]

        ### Then try breaking in y direction
        self._edges += [
        {
            "location": "check_decel_y",
            "guard": { "exp": Je.Eq(self.cardy, 0) },
            "destinations": [{
                "location": "send_acceleration",
            }]
        },
        {
            "location": "check_decel_y",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Ge(self.cardx, 0), Je.Gt(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_y",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Add(self.cardx, 1)), 2))),
                    Je.Value("map_dy", Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Sub(self.cardy, 1)), 2))),
                    Je.Value("ax", 0),
                    Je.Value("ay", -1),
                ],
            }]
        },
        {
            "location": "check_decel_y",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Ge(self.cardx, 0), Je.Lt(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_y",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Add(self.cardx, 1)), 2))),
                    Je.Value("map_dy", Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Add(self.cardy, 1)), 2)))),
                    Je.Value("ax", 0),
                    Je.Value("ay", 1),
                ],
            }]
        },
        {
            "location": "check_decel_y",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Lt(self.cardx, 0), Je.Gt(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_y",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Sub(self.cardx, 1)), 2)))),
                    Je.Value("map_dy", Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Sub(self.cardy, 1)), 2))),
                    Je.Value("ax", 0),
                    Je.Value("ay", -1),
                ],
            }]
        },
        {
            "location": "check_decel_y",
            "action": "check_if_valid",
            "guard": { "exp": Je.And(Je.Lt(self.cardx, 0), Je.Lt(self.cardy, 0)) },
            "destinations": [{
                "location": "await_decel_y",
                "assignments": [
                    Je.Value("map_x", self.carx),
                    Je.Value("map_y", self.cary),
                    Je.Value("map_dx", Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Sub(self.cardx, 1)), 2)))),
                    Je.Value("map_dy", Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Add(self.cardy, 1)), 2)))),
                    Je.Value("ax", 0),
                    Je.Value("ay", 1),
                ],
            }]
        },
        {
            "location": "await_decel_y",
            "action": "is_valid",
            "destinations": [{
                "location": "send_acceleration",
            }]
        },
        {
            "location": "await_decel_y",
            "action": "passed_finish_line",
            "destinations": [{
                "location": "send_acceleration",
            }]
        },
        ]

        # Otherwise break in both directions
        self._edges += [
        {
            "location": "await_decel_y",
            "action": "is_invalid",
            "guard": { "exp": 
                Je.And(Je.Le(self.cardx, 0), Je.Le(self.cardy, 0)) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", 1),
                    Je.Value("ay", 1),
                ],
            }]
        },
        {
            "location": "await_decel_y",
            "action": "is_invalid",
            "guard": { "exp": 
                Je.And(Je.Le(self.cardx, 0), Je.Gt(self.cardy, 0)) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", 1),
                    Je.Value("ay", -1),
                ],
            }]
        },
        {
            "location": "await_decel_y",
            "action": "is_invalid",
            "guard": { "exp": 
                Je.And(Je.Gt(self.cardx, 0), Je.Le(self.cardy, 0)) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", -1),
                    Je.Value("ay", 1),
                ],
            }]
        },
        {
            "location": "await_decel_y",
            "action": "is_invalid",
            "guard": { "exp": 
                Je.And(Je.Gt(self.cardx, 0), Je.Gt(self.cardy, 0)) },
            "destinations": [{
                "location": "send_acceleration",
                "assignments": [
                    Je.Value("ax", -1),
                    Je.Value("ay", -1),
                ],
            }]
        }
        ]


        ##############################################################
        ### Current velocity is valid
        ### -> try to accelerate towards goal

        # coordinate of closed goal:
        self._variables += [
                {"name": "goalx", "type": {"kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": self.grid.dimx}, "initial-value": 0 },
                {"name": "goaly", "type": {"kind": "bounded", "base": "int", "lower-bound": 0, "upper-bound": self.grid.dimy}, "initial-value": 0 },
        ]
        self._initial_values["goalx"] = 0
        self._initial_values["goaly"] = 0


        # if valid -> recompute closest goal coordinate
        _goal_coordinates = [(x, y) for x in range(self.grid.dimx) for y in range(self.grid.dimy) if self.grid.get(x, y) == CellType.GOAL]
        def _distance(i):
            return Je.Add(Je.Abs(Je.Sub(self.carx, _goal_coordinates[i][0])), Je.Abs(Je.Sub(self.cary, _goal_coordinates[i][1])))
        def _recursive_manhatten_min(i, j, ref):
            if j == len(_goal_coordinates):
                return _goal_coordinates[i][ref]
            if j == len(_goal_coordinates) - 1:
                return Je.Ite(Je.Le(_distance(i), _distance(j)), _goal_coordinates[i][ref], _goal_coordinates[j][ref])
            return Je.Ite(Je.Le(_distance(i), _distance(j)), _recursive_manhatten_min(i, j+1, ref), _recursive_manhatten_min(j, j+1, ref))
        self._edges.append({
            "location": "wait_result_current_velocity",
            "action": "is_valid",
            "destinations": [{
                "location": "current_velocity_valid",
                "assignments": [
                    Je.Value("goalx", _recursive_manhatten_min(0, 1, 0)),
                    Je.Value("goaly", _recursive_manhatten_min(0, 1, 1)),
                    Je.Value("dx", Je.Ite(Je.Le(self.cardx, 0), Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Sub(self.cardx, 1)), 2))), Je.Min(self.grid.dimx, Je.Div(Je.Mult(self.cardx, Je.Add(self.cardx, 1)), 2)))),
                    Je.Value("dy", Je.Ite(Je.Le(self.cardy, 0), Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Sub(self.cardy, 1)), 2))), Je.Min(self.grid.dimy, Je.Div(Je.Mult(self.cardy, Je.Add(self.cardy, 1)), 2)))),
                    Je.Value("dxm", Je.Ite(Je.Le(self.cardx, 1), Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(Je.Sub(self.cardx, 1), Je.Sub(self.cardx, 2)), 2))), Je.Min(self.grid.dimx, Je.Div(Je.Mult(Je.Sub(self.cardx, 1), self.cardx), 2)))),
                    Je.Value("dym", Je.Ite(Je.Le(self.cardy, 1), Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(Je.Sub(self.cardy, 1), Je.Sub(self.cardy, 2)), 2))), Je.Min(self.grid.dimy, Je.Div(Je.Mult(Je.Sub(self.cardy, 1), self.cardy), 2)))),
                    Je.Value("dxp", Je.Ite(Je.Lt(self.cardx, -1), Je.Neg(Je.Min(self.grid.dimx, Je.Div(Je.Mult(Je.Add(self.cardx, 1), self.cardx), 2))), Je.Min(self.grid.dimx, Je.Div(Je.Mult(Je.Add(self.cardx, 1), Je.Add(self.cardx, 2)), 2)))),
                    Je.Value("dyp", Je.Ite(Je.Lt(self.cardy, -1), Je.Neg(Je.Min(self.grid.dimy, Je.Div(Je.Mult(Je.Add(self.cardy, 1), self.cardy), 2))), Je.Min(self.grid.dimy, Je.Div(Je.Mult(Je.Add(self.cardy, 1), Je.Add(self.cardy, 2)), 2)))),
                ],
            }]
        })

        # update x accel
        self._edges += [
            {
                "location": "current_velocity_valid",
                "guard": { "exp": Je.And(Je.Eq(Je.Add(self.carx, self.cardx), "goalx"), Je.Gt(Je.Abs(self.cardx), 1)) },
                "destinations": [{
                    "location": "send_acceleration",
                    "assignments": [
                        Je.Value("ax", Je.Ite(Je.Lt(self.cardx, 0), 1, -1)),
                    ],
                }]
            },
            {
                "location": "current_velocity_valid",
                "guard": { "exp": Je.And(Je.Eq(Je.Add(self.carx, self.cardx), "goalx"), Je.Le(Je.Abs(self.cardx), 1)) },
                "destinations": [{
                    "location": "determine_accel_y",
                }]
            },
            {
                "location": "current_velocity_valid",
                "guard": { "exp": Je.Lt(Je.Add(self.carx, self.cardx), "goalx") },
                "action": "check_if_valid",
                "destinations": [{
                    "location": "check_accel_x",
                    "assignments": [
                        Je.Value("map_x", self.carx),
                        Je.Value("map_y", self.cary),
                        Je.Value("map_dx", "dxp"),
                        Je.Value("map_dy", "dy"),
                        Je.Value("ax", 1),
                    ]
                }]
            },
            {
                "location": "current_velocity_valid",
                "guard": { "exp": Je.Gt(Je.Add(self.carx, self.cardx), "goalx") },
                "action": "check_if_valid",
                "destinations": [{
                    "location": "check_accel_x",
                    "assignments": [
                        Je.Value("map_x", self.carx),
                        Je.Value("map_y", self.cary),
                        Je.Value("map_dx", "dxm"),
                        Je.Value("map_dy", self.cardy),
                        Je.Value("ax", -1),
                    ]
                }]
            },
            {
                "location": "check_accel_x",
                "action": "passed_finish_line",
                "destinations": [{
                    "location": "send_acceleration",
                }]
            },
            {
                "location": "check_accel_x",
                "action": "is_invalid",
                "destinations": [{
                    "location": "determine_accel_y",
                    "assignments": [
                        Je.Value("ax", 0),
                    ]
                }]
            },
            {
                "location": "check_accel_x",
                "action": "is_valid",
                "guard": { "exp": Je.Ne(self.cardy, 0) },
                "destinations": [{
                    "location": "send_acceleration",
                }]
            },
            {
                "location": "check_accel_x",
                "action": "is_valid",
                "guard": { "exp": Je.And(Je.Eq(self.cardy, 0), Je.Eq("ax", -1)) },
                "destinations": [{
                    "location": "determine_accel_y",
                    "assignments": [
                        Je.Value("dx", "dxm"),
                    ]
                }]
            },
            {
                "location": "check_accel_x",
                "action": "is_valid",
                "guard": { "exp": Je.And(Je.Eq(self.cardy, 0), Je.Eq("ax", 1)) },
                "destinations": [{
                    "location": "determine_accel_y",
                    "assignments": [
                        Je.Value("dx", "dxp"),
                    ]
                }]
            }
        ]

        # update y
        self._edges += [
            {
                "location": "determine_accel_y",
                "guard": { "exp": Je.Eq(Je.Add(self.cary, self.cardy), "goaly") },
                "destinations": [{
                    "location": "send_acceleration",
                }]
            },
            {
                "location": "determine_accel_y",
                "guard": { "exp": Je.Lt(Je.Add(self.cary, self.cardy), "goaly") },
                "action": "check_if_valid",
                "destinations": [{
                    "location": "check_accel_y",
                    "assignments": [
                        Je.Value("map_x", self.carx),
                        Je.Value("map_y", self.cary),
                        Je.Value("map_dx", "dx"),
                        Je.Value("map_dy", "dyp"),
                        Je.Value("ay", 1),
                    ]
                }]
            },
            {
                "location": "determine_accel_y",
                "guard": { "exp": Je.Gt(Je.Add(self.cary, self.cardy), "goaly") },
                "action": "check_if_valid",
                "destinations": [{
                    "location": "check_accel_y",
                    "assignments": [
                        Je.Value("map_x", self.carx),
                        Je.Value("map_y", self.cary),
                        Je.Value("map_dx", "dx"),
                        Je.Value("map_dy", "dym"),
                        Je.Value("ay", -1),
                    ]
                }]
            },
            {
                "location": "check_accel_y",
                "action": "passed_finish_line",
                "destinations": [{
                    "location": "send_acceleration",
                }]
            },
            {
                "location": "check_accel_y",
                "action": "is_invalid",
                "destinations": [{
                    "location": "send_acceleration",
                    "assignments": [
                        Je.Value("ay", 0),
                    ]
                }]
            },
            {
                "location": "check_accel_y",
                "action": "is_valid",
                "destinations": [{
                    "location": "send_acceleration",
                }]
            }
        ]

    def jani(self):
        return {
            "name": "car%d_controller" % self.carid,
            "locations": [ { "name": loc } for loc in sorted(self._locations) ],
            "variables": self._variables,
            "restrict-initial": { "exp": Je.And([Je.Eq(x, self._initial_values[x]) for x in sorted(self._initial_values.keys())]) },
            "initial-locations": [ self._initial_loc ],
            "edges": self._edges,
        }

    def __str__(self):
        return json.dumps(self.jani(), ensure_ascii=False, sort_keys=True, indent=2)


def controller(grid, carid):
    return CarControllerAutomaton(grid, carid)


